"""
The `dbinstance` module creates a singleton `MatrixDB` database
instance, populating it from ssstats.csv if necessary.
"""
import datetime
import logging

from . import csvindex
from .db import MatrixDB

logger = logging.getLogger(__name__)


instance = MatrixDB()

if instance.nrows == 0 or (
    datetime.datetime.utcnow() - instance.last_update
) > datetime.timedelta(days=90):
    logger.info("{Re}creating index from CSV file...")
    instance.refresh(csvindex.generate())



import argparse
import logging
import sys

from .config import SS_DIR
from .dbinstance import instance

logger = logging.getLogger(__name__)


def search(name_or_id=None, **kwargs):
    """
    Search for matrix/matrices with a given name pattern or numeric ID.
    Optionally, limit search to matrices of a specific data type or
    with the specified range of rows, columns and non-zero values.
    """
    logger.debug("Name or ID = " + str(name_or_id))
    if name_or_id is not None:
        if isinstance(name_or_id, str):
            if "/" in name_or_id:
                group, name = name_or_id.split("/")
                kwargs["group"] = group
                if not name == "" and not name == "*":
                    kwargs["name"] = name
            else:
                kwargs["name"] = name_or_id
        elif isinstance(name_or_id, int):
            kwargs["matid"] = name_or_id
        else:
            raise ValueError(
                "First argument to search " + "must be a string or an integer"
            )

    return instance.search(**kwargs)


def fetch(
    name_or_id=None, format="MM", location=None, dry_run=False, **kwargs
):
    matrices = search(name_or_id, **kwargs)
    if len(matrices) > 0:
        logger.info(
            "Found %d %s"
            % (len(matrices), "entry" if len(matrices) == 1 else "entries")
        )
        for matrix in matrices:
            logger.info(
                "Downloading %s/%s to %s"
                % (
                    matrix.group,
                    matrix.name,
                    matrix.localpath(format, location, extract=True)[0],
                )
            )
            if not dry_run:
                matrix.download(format, location, extract=True)
    return matrices


def cli(argv=sys.argv[1:]):
    parser = argparse.ArgumentParser(prog="ssget")

    parser.add_argument(
        "-i",
        "--id",
        action="store",
        type=int,
        dest="matid",
        help="Download a matrix with the given ID.",
    )
    parser.add_argument(
        "-g",
        "--group",
        action="store",
        type=str,
        dest="group",
        help="The matrix group.",
    )
    parser.add_argument(
        "-n",
        "--name",
        action="store",
        type=str,
        dest="name",
        help="The name or a pattern matching the name of the matrix/matrices.",
    )
    parser.add_argument(
        "-d",
        "--data-type",
        action="store",
        type=str,
        dest="dtype",
        help="The element type of the matrix/matrices"
        ", can be one of 'real', 'complex' or 'binary'.",
    )
    parser.add_argument(
        "-f",
        "--format",
        action="store",
        type=str,
        dest="format",
        default="MM",
        help="The format in which to download the matrix/matrices.\
              Can be one of 'MM', 'MAT' or 'RB' for MatrixMarket, \
              MATLAB or Rutherford-Boeing formats respectively.",
    )
    parser.add_argument(
        "-l",
        "--limit",
        action="store",
        type=int,
        default=10,
        dest="limit",
        help="The maximum number of matrices to be downloaded.",
    )
    parser.add_argument(
        "-o",
        "--outdir",
        action="store",
        type=str,
        dest="location",
        help="The directory in the local machine where matrices will be \
              downloaded to. Defaults to "
        + SS_DIR,
    )
    parser.add_argument(
        "--dry-run",
        action="store_true",
        dest="dry_run",
        default=False,
        help="If True, only print the matrices that will be downloaded \
              but do not actually download them.",
    )

    g = parser.add_argument_group(
        "Size and Non-zero filters",
        "These options may be used to restrict the shape or number "
        + "of non-zero elements of the matrices to be downloaded",
    )

    g.add_argument(
        "--min-rows",
        action="store",
        type=int,
        dest="min_rows",
        help="The minimum number of rows in the matrix/matrices.",
    )
    g.add_argument(
        "--max-rows",
        action="store",
        type=int,
        dest="max_rows",
        help="The maximum number of rows in the matrix/matrices.",
    )
    g.add_argument(
        "--min-cols",
        action="store",
        type=int,
        dest="min_cols",
        help="The minimum number of columns in the matrix/matrices.",
    )
    g.add_argument(
        "--max-cols",
        action="store",
        type=int,
        dest="max_cols",
        help="The maximum number of columns in the matrix/matrices.",
    )
    g.add_argument(
        "--min-nnzs",
        action="store",
        type=int,
        dest="min_nnzs",
        help="The minimum number of non-zero values in the matrix/matrices.",
    )
    g.add_argument(
        "--max-nnzs",
        action="store",
        type=int,
        dest="max_nnzs",
        help="The maximum number of non-zero values in the matrix/matrices.",
    )

    lg = parser.add_argument_group(
        "Logging and verbosity options",
        "These options govern the level of spew from ssgetpy. "
        + "By default, ssgetpy prints a small number of messages "
        + "such as the number of matrices being downloaded and "
        + "where they are being downloaded to. "
        + "To suppress these message, pass --quiet. "
        + "To enable debug diagnostics, pass --verbose.",
    )
    lg.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        dest="verbose",
        default=False,
        help="Enable debug diagnostics.",
    )
    lg.add_argument(
        "-q",
        "--quiet",
        action="store_true",
        dest="quiet",
        default=False,
        help="Do not print any messages to the console.",
    )

    if len(argv) == 0:
        parser.print_help()
        return

    args = parser.parse_args(argv)

    optdict = dict(
        matid=args.matid,
        group=args.group,
        name=args.name,
        rowbounds=(args.min_rows, args.max_rows),
        colbounds=(args.min_cols, args.max_cols),
        nzbounds=(args.min_nnzs, args.max_nnzs),
        dtype=args.dtype,  # isspd     = args.isspd,\
        limit=args.limit,
    )

    if args.quiet:
        pass
    elif args.verbose:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    name_or_id = None
    if args.matid:
        name_or_id = args.matid
    elif args.name:
        name_or_id = args.matid

    fetch(name_or_id, args.format, args.location, args.dry_run, **optdict)



import logging
import os
import sys

logger = logging.getLogger(__name__)

SS_DIR = None
SS_DB = "index.db"
SS_TABLE = "MATRICES"
SS_ROOT_URL = "https://sparse.tamu.edu"
SS_INDEX_URL = "/".join((SS_ROOT_URL, "files", "ssstats.csv"))

if sys.platform == "win32":
    SS_DIR = os.path.join(os.environ["APPDATA"], "ssgetpy")
else:
    SS_DIR = os.path.join(os.environ["HOME"], ".ssgetpy")

SS_DB = os.path.join(SS_DIR, SS_DB)

os.makedirs(SS_DIR, exist_ok=True)


def dump():
    logger.debug(
        dict(
            SS_DIR=SS_DIR,
            SS_DB=SS_DB,
            SS_TABLE=SS_TABLE,
            SS_ROOT_URL=SS_ROOT_URL,
            SS_INDEX_URL=SS_INDEX_URL,
        )
    )



import datetime
import logging

from .config import SS_DB, SS_TABLE
from .matrix import Matrix, MatrixList

logger = logging.getLogger(__name__)


def _from_timestamp(timestamp):
    if hasattr(datetime.datetime, "fromisoformat"):
        return datetime.datetime.fromisoformat(timestamp)
    return datetime.datetime.strptime(timestamp, "%Y-%m-%d %H:%M:%S")


class MatrixDB:
    def __init__(self, db=SS_DB, table=SS_TABLE):
        import sqlite3

        self.db = db
        self.matrix_table = table
        self.update_table = "update_table"
        self.conn = sqlite3.connect(self.db)
        self._create_table()

    def _get_nrows(self):
        return int(
            self.conn.execute(
                "SELECT COUNT(*) FROM %s" % self.matrix_table
            ).fetchall()[0][0]
        )

    nrows = property(_get_nrows)

    def _get_last_update(self):
        last_update = self.conn.execute(
            "SELECT MAX(update_date) " + f"from {self.update_table}"
        ).fetchall()[0][0]
        return (
            _from_timestamp(last_update)
            if last_update
            else datetime.datetime.utcfromtimestamp(0)
        )

    last_update = property(_get_last_update)

    def _drop_table(self):
        self.conn.execute("DROP TABLE IF EXISTS %s" % self.matrix_table)
        self.conn.execute(f"DROP TABLE IF EXISTS {self.update_table}")
        self.conn.commit()

    def _create_table(self):
        self.conn.execute(
            """CREATE TABLE IF NOT EXISTS %s (
                             id INTEGER PRIMARY KEY,
                             matrixgroup TEXT,
                             name TEXT,
                             rows INTEGER,
                             cols INTEGER,
                             nnz INTEGER,
                             dtype TEXT,
                             is2d3d INTEGER,
                             isspd INTEGER,
                             psym REAL,
                             nsym REAL,
                             kind TEXT)"""
            % self.matrix_table
        )

        self.conn.execute(
            f"CREATE TABLE IF NOT EXISTS {self.update_table} "
            + "(update_date TIMESTAMP)"
        )
        self.conn.commit()

    def insert(self, values):
        self.conn.executemany(
            "INSERT INTO %s VALUES(?,?,?,?,?,?,?,?,?,?,?,?)"
            % self.matrix_table,
            values,
        )
        self.conn.execute(
            f"INSERT INTO {self.update_table} " + "VALUES (datetime('now'))"
        )
        self.conn.commit()

    def refresh(self, values):
        self._drop_table()
        self._create_table()
        self.insert(values)

    def dump(self):
        return self.conn.execute(
            "SELECT * from %s" % self.matrix_table
        ).fetchall()

    @staticmethod
    def _is_constraint(field, value):
        return value and "(%s = '%s')" % (field, value)

    @staticmethod
    def _like_constraint(field, value):
        return value and "(%s LIKE '%%%s%%')" % (field, value)

    @staticmethod
    def _sz_constraint(field, bounds):
        if bounds is None or (bounds[0] is None and bounds[1] is None):
            return None
        constraints = []
        if bounds[0] is not None:
            constraints.append("%s >= %d" % (field, bounds[0]))
        if bounds[1] is not None:
            constraints.append("%s <= %d" % (field, bounds[1]))
        return " ( " + " AND ".join(constraints) + " ) "

    @staticmethod
    def _bool_constraint(field, value):
        if value is None:
            return None
        elif value:
            return "(%s = 1)" % field
        else:
            return "(%s = 0)" % field

    def search(
        self,
        matid=None,
        group=None,
        name=None,
        rowbounds=None,
        colbounds=None,
        nzbounds=None,
        dtype=None,
        is2d3d=None,
        isspd=None,
        kind=None,
        limit=10,
    ):

        querystring = "SELECT * FROM %s" % self.matrix_table

        mid_constraint = MatrixDB._is_constraint("id", matid)
        grp_constraint = MatrixDB._is_constraint("matrixgroup", group)
        nam_constraint = MatrixDB._like_constraint("name", name)
        row_constraint = MatrixDB._sz_constraint("rows", rowbounds)
        col_constraint = MatrixDB._sz_constraint("cols", colbounds)
        nnz_constraint = MatrixDB._sz_constraint("nnz", nzbounds)
        dty_constraint = MatrixDB._is_constraint("dtype", dtype)
        geo_constraint = MatrixDB._bool_constraint("is2d3d", is2d3d)
        spd_constraint = MatrixDB._bool_constraint("isspd", isspd)
        knd_constraint = MatrixDB._like_constraint("kind", kind)

        constraints = list(
            filter(
                lambda x: x is not None,
                (
                    mid_constraint,
                    grp_constraint,
                    nam_constraint,
                    row_constraint,
                    col_constraint,
                    nnz_constraint,
                    dty_constraint,
                    geo_constraint,
                    spd_constraint,
                    knd_constraint,
                ),
            )
        )

        if any(constraints):
            querystring += " WHERE " + " AND ".join(constraints)

        querystring += " LIMIT (%s)" % limit

        logger.debug(querystring)

        return MatrixList(
            Matrix(*x) for x in self.conn.execute(querystring).fetchall()
        )



"""
The `ssgetpy` module provides interfaces to search and download matrices
from the SuiteSparse Matrix Collection.

There are two ways to use `ssgetpy`:
* By importing the `ssgetpy` module in Python, or
* As a standalone command-line tool

To search for matrices that match a given criterion, use `ssgetpy.search`:

`ssgetpy.search(name_or_id, **kwargs)`

`ssgetpy.search` only returns a list of `Matrix` objects that match
the selection criterion. To download the matrices themselves, use the
`download` method in the `Matrix object` or use `ssgetpy.fetch` ::

  `ssgetpy.fetch(name_or_id, format, location, **kwargs)`

The rules for specifying the search criteria in `ssgetpy.search` and
 `ssgetpy.fetch` are as follows:

1. `name_or_id` can be either the numerical ID of the matrix such as
   `42` or a pattern such as `"HB/ash*"` or `"c-"`. This field is
   optional.
2. `**kwargs` is a set of key-value pairs with search constraints:
     - matid: An integer matrix ID such as `42`
     - group: The matrix group name such as `HB`
     - name: A pattern containing the matrix name such as `ash-`
     - rowbounds: A tuple of the form (min, max) containing the
       minimum and maximum rows. The min or max value can be set to
       `None`.
     - colbounds: A tuple of the form (min, max) containing the minimum
                  and maximum columns.
     - nzbounds: A tuple of the form (min, max) containing the minimum
                 and maximum non-zero values.
     - dtype: The matrix data type, one of `real`, `complex` or `binary`.
     - is2d3d: If true, only selects matrices arising from
               2D and 3D discretizations.
     - isspd: If true, only selects SPD matrices.
     - kind: A string describing the problem domain,
             see http://www.cise.ufl.edu/research/sparse/matrices/kind.html
     - limit: Number of matrices to return, defaults to 10.

If `name_or_id` is specified, it overrides any conflicting key-value settings
in `**kwargs`.

In `ssgetpy.fetch`, `format` can be one of 'MM', 'MAT' or 'RB'; 'MM'
is the default if `format` is omitted.  Finally, `location` refers
to the directory where the matrices will be downloaded on the local
machine. It defaults to `%APPDATA%/`ssgetpy`` on Windows and
`~/.ssgetpy` on Unix-like platforms.

In addition to its usage as a Python library, `ssgetpy` can be run from
the command line as follows ::

  python `ssgetpy`
  Usage: `ssgetpy` [options]

  Options:
    -h, --help            show this help message and exit
    -i MATID, --id=MATID  Download a matrix with the given ID.
    -g GROUP, --group=GROUP
                          The matrix group.
    -n NAME, --name=NAME  The name or a pattern matching the name of the
                          matrix/matrices.
    -d DTYPE, --data-type=DTYPE
                          The element type of the matrix/matrices,
                          can be one of 'real', 'complex' or 'binary'.
    -s, --spd             Only selects SPD matrices.
    -f FORMAT, --format=FORMAT
                          The format in which to download the matrix/matrices.
                          Can be one of 'MM', 'MAT' or 'RB' for MatrixMarket,
                          MATLAB or Rutherford-Boeing formats respectively.
                          Defaults to 'MM'.
    -l LIMIT, --limit=LIMIT
                          The maximum number of matrices to be downloaded.
                          Defaults to 10.
    -o LOCATION, --outdir=LOCATION
                          The directory in the local machine where matrices
                          will be downloaded to.
                          Defaults to `%AppData%/ssgetpy` on Windows
                          and `~/.ssgetpy` on Unix.

    Size and Non-zero filters:
      These options may be used to restrict the shape or number of non-zero
      elements of the matrices to be downloaded

      --min-rows=MIN_ROWS
                          The minimum number of rows in the matrix/matrices.
      --max-rows=MAX_ROWS
                          The maximum number of rows in the matrix/matrices.
      --min-cols=MIN_COLS
                          The minimum number of columns in the matrix/matrices.
      --max-cols=MAX_COLS
                          The maximum number of columns in the matrix/matrices.
      --min-nnzs=MIN_NNZS
                          The minimum number of non-zero values in the
                          matrix/matrices.
      --max-nnzs=MAX_NNZS
                          The maximum number of non-zero values in the
                          matrix/matrices.
"""
from .query import fetch, search

__all__ = ["fetch", "search"]



import gzip
import os
import shutil
import tarfile


def extract(bundle):
    basedir, filename = os.path.split(bundle)
    tarfilename = os.path.join(
        basedir, ".".join((filename.split(".")[0], "tar"))
    )
    gzfile = gzip.open(bundle, "rb")
    with open(tarfilename, "wb") as outtarfile:
        shutil.copyfileobj(gzfile, outtarfile)
    gzfile.close()
    tarfile.open(tarfilename).extractall(basedir)
    os.unlink(tarfilename)
    os.unlink(bundle)



import os
import time

import requests
from tqdm.auto import tqdm

from . import bundle
from .config import SS_DIR, SS_ROOT_URL


class MatrixList(list):
    def _repr_html_(self):
        body = "".join(r.to_html_row() for r in self)
        return f"<table>{Matrix.html_header()}<tbody>{body}</tbody></table>"

    def __getitem__(self, expr):
        result = super().__getitem__(expr)
        return MatrixList(result) if isinstance(expr, slice) else result

    def download(self, format="MM", destpath=None, extract=False):
        with tqdm(total=len(self), desc="Overall progress") as pbar:
            for matrix in self:
                matrix.download(format, destpath, extract)
                pbar.update(1)


class Matrix:
    """
    A `Matrix` object represents an entry in the SuiteSparse matrix collection.
    It has the following attributes:
    `id`   : The unique identifier for the matrix in the database.
    `group`: The name of the group this matrix belongs to.
    `name` : The name of this matrix.
    `rows` : The number of rows.
    `cols` : The number of columns.
    `nnz`  : The number of non-zero elements.
    `dtype`: The datatype of non-zero elements: `real`, `complex` or `binary`
    `is2d3d`: True if this matrix comes from a 2D or 3D discretization.
    `isspd` : True if this matrix is symmetric, positive definite
    `kind`  : The underlying problem domain
    """

    attr_list = [
        "Id",
        "Group",
        "Name",
        "Rows",
        "Cols",
        "NNZ",
        "DType",
        "2D/3D Discretization?",
        "SPD?",
        "Pattern Symmetry",
        "Numerical Symmetry",
        "Kind",
        "Spy Plot",
    ]

    @staticmethod
    def html_header():
        return (
            "<thead>"
            + "".join(f"<th>{attr}</th>" for attr in Matrix.attr_list)
            + "</thead>"
        )

    def __init__(
        self,
        identifier,
        group,
        name,
        rows,
        cols,
        nnz,
        dtype,
        is2d3d,
        isspd,
        psym,
        nsym,
        kind,
    ):
        self.id = identifier
        self.group = group
        self.name = name
        self.rows = rows
        self.cols = cols
        self.nnz = nnz
        self.dtype = dtype
        self.is2d3d = not not is2d3d
        self.isspd = not not isspd
        self.psym = psym
        self.nsym = nsym
        self.kind = kind

    def to_tuple(self):
        """
        Returns the fields in a `Matrix` instance in a tuple.
        """
        return (
            self.id,
            self.group,
            self.name,
            self.rows,
            self.cols,
            self.nnz,
            self.dtype,
            self.is2d3d,
            self.isspd,
            self.psym,
            self.nsym,
            self.kind,
            self.icon_url(),
        )

    def _render_item_html(self, key, value):
        if key == "Spy Plot":
            return f'<img src="{value}">'
        if key == "Group":
            return f'<a href="{self.group_info_url()}" target="_blank">{value}</a>'
        if key == "Name":
            return f'<a href="{self.matrix_info_url()}" target="_blank">{value}</a>'
        if key in ("Pattern Symmetry", "Numerical Symmetry"):
            return f"{value:0.2}"
        if key in ("2D/3D Discretization?", "SPD?"):
            return "Yes" if value else "No"

        return str(value)

    def to_html_row(self):
        return (
            "<tr>"
            + "".join(
                f"<td>{self._render_item_html(key, value)}</td>"
                for key, value in zip(Matrix.attr_list, self.to_tuple())
            )
            + "</tr>"
        )

    def _filename(self, format="MM"):
        if format == "MM" or format == "RB":
            return self.name + ".tar.gz"
        elif format == "MAT":
            return self.name + ".mat"
        else:
            raise ValueError("Format must be 'MM', 'MAT' or 'RB'")

    def _defaultdestpath(self, format="MM"):
        return os.path.join(SS_DIR, format, self.group)

    def icon_url(self):
        return "/".join((SS_ROOT_URL, "files", self.group, self.name + ".png"))

    def group_info_url(self):
        return "/".join((SS_ROOT_URL, self.group))

    def matrix_info_url(self):
        return "/".join((SS_ROOT_URL, self.group, self.name))

    def url(self, format="MM"):
        """
        Returns the URL for this `Matrix` instance.
        """
        fname = self._filename(format)
        directory = format.lower() if format == "MAT" else format
        return "/".join((SS_ROOT_URL, directory, self.group, fname))

    def localpath(self, format="MM", destpath=None, extract=False):
        destpath = destpath or self._defaultdestpath(format)

        # localdestpath is the directory containing the unzipped files
        # in the case of MM and RB (if extract is true) or
        # the file itself in the case of MAT (or if extract is False)
        localdest = os.path.join(destpath, self._filename(format))
        localdestpath = (
            localdest
            if (format == "MAT" or not extract)
            else os.path.join(destpath, self.name)
        )

        return localdestpath, localdest

    def download(self, format="MM", destpath=None, extract=False):
        """
        Downloads this `Matrix` instance to the local machine,
        optionally unpacking any TAR.GZ files.
        """
        # destpath is the directory containing the matrix
        # It is of the form ~/.PyUFGet/MM/HB
        destpath = destpath or self._defaultdestpath(format)

        # localdest is matrix file (.MAT or .TAR.GZ)
        # if extract = True, localdestpath is the directory
        # containing the unzipped matrix
        localdestpath, localdest = self.localpath(format, destpath, extract)

        if not os.access(localdestpath, os.F_OK):
            # Create the destination path if necessary
            os.makedirs(destpath, exist_ok=True)

            response = requests.get(self.url(format), stream=True)
            content_length = int(response.headers["content-length"])

            with open(localdest, "wb") as outfile, tqdm(
                total=content_length, desc=self.name, unit="B"
            ) as pbar:
                for chunk in response.iter_content(chunk_size=4096):
                    outfile.write(chunk)
                    pbar.update(4096)
                    time.sleep(0.1)

            if extract and (format == "MM" or format == "RB"):
                bundle.extract(localdest)

        return localdestpath, localdest

    def __str__(self):
        return str(self.to_tuple())

    def __repr__(self):
        return "Matrix" + str(self.to_tuple())

    def _repr_html_(self):
        return (
            f"<table>{Matrix.html_header()}"
            + f"<tbody>{self.to_html_row()}</tbody></table>"
        )



import pathlib

from setuptools import setup

HERE = pathlib.Path(__file__).parent

README = (HERE / "README.md").read_text()

setup(
    name="ssgetpy",
    version="1.0-pre2",
    description="A Python interface to the SuiteSparse Matrix Collection",
    author="Sudarshan Raghunathan",
    author_email="darshan@alum.mit.edu",
    url="http://www.github.com/drdarshan/ssgetpy",
    long_description=README,
    long_description_content_type="text/markdown",
    packages=["ssgetpy"],
    entry_points={"console_scripts": ["ssgetpy = ssgetpy.query:cli", ], },
    python_requires=">3.5.2",
    install_requires=["requests>=2.22", "tqdm>=4.41"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Mathematics",
    ],
)



"""
The `csvindex` module parses the SSStats.csv file and
generates entries for each row in a Matrix database
"""

import csv
import logging

import requests

from .config import SS_INDEX_URL

logger = logging.getLogger(__name__)


def getdtype(real, logical):
    """
    Converts a (real, logical) pair into one of the three types:
    'real', 'complex' and 'binary'
    """
    return "binary" if logical else ("real" if real else "complex")


def gen_rows(csvrows):
    """
    Creates a generator that returns a single row in the matrix database.
    """
    reader = csv.reader(csvrows)
    matid = 0
    for line in reader:
        matid += 1
        group = line[0]
        name = line[1]
        rows = int(line[2])
        cols = int(line[3])
        nnz = int(line[4])
        real = bool(int(line[5]))
        logical = bool(int(line[6]))
        is2d3d = bool(int(line[7]))
        isspd = bool(int(line[8]))
        psym = float(line[9])
        nsym = float(line[10])
        kind = line[11]
        yield matid, group, name, rows, cols, nnz, getdtype(
            real, logical
        ), is2d3d, isspd, psym, nsym, kind


def generate():
    response = requests.get(SS_INDEX_URL)
    lines = response.iter_lines()

    # Read the number of entries
    logger.info(f"Number of entries in the CSV file: {next(lines)}")
    # Read the last modified date
    logger.info(f"Last modified date: {next(lines)}")

    return gen_rows(line.decode("utf-8") for line in lines)
